@import url("bootstrap-astropy.css");

div.topbar a.brand {
    background: none;
    background-image: none;
}

div.topbar ul li a.homelink {
    background: none;
    background-image: none;
}

div.topbar a.brand #logotext1{
    font-weight: 200;
    padding-right: 4px;
}

div.topbar a.brand #logotext2{
    font-weight: 600;
}
